import logging
l = logging.getLogger("claripy.frontends.composite_frontend")

import itertools
symbolic_count = itertools.count()

from .caching_frontend import CachingFrontend

class CompositeFrontend(CachingFrontend):
	def __init__(self, template_frontend, **kwargs):
		CachingFrontend.__init__(self, **kwargs)
		self._solvers = { }
		self._template_frontend = template_frontend

		self._solvers['CONCRETE'] = self._merged_solver_for({'CONCRETE'})

	#
	# Serialization stuff
	#

	def _ana_getstate(self):
		return self._solvers, self._template_frontend, CachingFrontend._ana_getstate(self)

	def _ana_setstate(self, s):
		self._solvers, self._template_frontend, base_state = s
		CachingFrontend._ana_setstate(base_state)

	#
	# Frontend management
	#

	@property
	def _solver_list(self):
		seen_solvers = set()
		solver_list = [ ]
		for s in self._solvers.itervalues():
			if id(s) in seen_solvers: continue
			seen_solvers.add(id(s))
			solver_list.append(s)
		return solver_list

	@property
	def variables(self):
		return set(self._solvers.keys()) - { 'CONCRETE' }

	@property
	def constraints(self):
		return sum([ s.constraints for s in self._solver_list ], [ ])

	def _solvers_for_variables(self, names):
		seen_solvers = set()
		existing_solvers = [ ]
		for n in names:
			if n not in self._solvers: continue
			s = self._solvers[n]

			if id(s) in seen_solvers: continue
			seen_solvers.add(id(s))
			existing_solvers.append(s)
		return existing_solvers

	def _merged_solver_for(self, names):
		l.debug("composite_solver._merged_solver_for() running with %d names", len(names))
		solvers = self._solvers_for_variables(names)
		if len(solvers) == 0:
			l.debug("... creating new solver")
			return self._template_frontend._blank_copy()
		elif len(solvers) == 1:
			l.debug("... got one solver")
			return solvers[0]
		else:
			l.debug(".... combining %d solvers", len(solvers))
			return solvers[0].combine(solvers[1:])

	def _shared_solvers(self, others):
		"""
		Returns a sequence of the solvers that self and others share.
		"""

		solvers_by_id = { s.uuid: s for s in self._solver_list }
		common_solvers = set(solvers_by_id.keys())
		other_sets = [ { s.uuid for s in cs._solver_list } for cs in others ]
		for o in other_sets: common_solvers &= o

		return [ solvers_by_id[s] for s in common_solvers ]

	def _variable_sets(self):
		return { s.variables for s in self._solver_list }

	def _shared_varsets(self, others):
		common_varsets = self._variable_sets()
		for o in others: common_varsets &= o.all_varsets()
		return common_varsets

	#
	# Constraints
	#

	def _add_dependent_constraints(self, names, constraints):
		l.debug("Adding %d constraints to %d names", len(constraints), len(names))
		s = self._merged_solver_for(names)
		s.add(constraints)
		for v in s.variables | names:
			self._solvers[v] = s

	def _add_constraints(self, constraints, invalidate_cache=True):
		if not invalidate_cache:
			l.warning("ignoring non-invalidating constraints")
			return

		split = self._split_constraints(constraints)

		l.debug("%s, solvers before: %d", self, len(self._solvers))
		for names,set_constraints in split:
			self._add_dependent_constraints(names, set_constraints)
		l.debug("... solvers after add: %d", len(self._solver_list))

	#
	# Solving
	#

	def _solve(self, extra_constraints=(), exact=None, cache=None):
		l.debug("%r checking satisfiability...", self)

		if len(extra_constraints) != 0:
			extra_vars = frozenset.union(*(a.variables for a in extra_constraints))
			solvers = [ self._merged_solver_for(extra_vars) ]
			for s in self._solver_list:
				if len(s.variables | solvers[0].variables) == 0:
					solvers.append(s)
		else:
			solvers = self._solver_list

		model = { }
		satness = True

		for s in solvers:
			if not s.satisfiable(extra_constraints=extra_constraints if s is solvers[0] else (), exact=exact, cache=cache):
				l.debug("... %r: False", s)
				satness = False
				break

			l.debug("... %r: True", s)
			model.update(s.result.model)

		l.debug("... ok!")
		return Result(satness, model=model)

	def _all_variables(self, e, extra_constraints=()): #pylint:disable=no-self-use
		all_vars = e.variables
		if len(extra_constraints) != 0:
			all_vars |= frozenset.union(*(a.variables for a in extra_constraints))
		if len(all_vars) == 0:
			all_vars = { 'CONCRETE' }
		return all_vars

	def _eval(self, e, n, extra_constraints=(), exact=None, cache=None):
		return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).eval(e, n, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _batch_eval(self, exprs, n, extra_constraints=(), exact=None, cache=None):

		the_solver = None
		for expr in exprs:
			s = self._merged_solver_for(self._all_variables(expr, extra_constraints=extra_constraints))
			if the_solver is None:
				the_solver = s
			elif the_solver is not s:
				raise ClaripyFrontendError("having expressions across multiple solvers is not supported by _batch_eval() right now")

		return the_solver.batch_eval(exprs, n, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _max(self, e, extra_constraints=(), exact=None, cache=None):
		return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).max(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _min(self, e, extra_constraints=(), exact=None, cache=None):
		return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).min(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _solution(self, e, n, extra_constraints=(), exact=None, cache=None):
		return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).solution(e, n, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _is_true(self, e, extra_constraints=(), exact=None, cache=None):
		return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).is_true(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _is_false(self, e, extra_constraints=(), exact=None, cache=None):
		return self._merged_solver_for(self._all_variables(e, extra_constraints=extra_constraints)).is_false(e, extra_constraints=extra_constraints, exact=exact, cache=cache)

	def _simplify(self):
		l.debug("Simplifying %r with %d solvers", self, len(self._solver_list))
		for s in self._solver_list:
			if s._simplified:
				continue

			l.debug("... simplifying child solver %r", s)
			s.simplify()
			l.debug("... splitting child solver %r", s)
			split = s.split()

			l.debug("... split solver %r into %d parts", s, len(split))
			l.debug("... variable counts: %s", [ len(ss.variables) for ss in split ])
			if len(split) > 1:
				for s in split:
					for v in s.variables:
						self._solvers[v] = s
		l.debug("... after-split, %r has %d solvers", self, len(self._solver_list))

	#
	# Merging and splitting
	#

	def finalize(self):
		l.error("CompositeFrontend.finalize is incomplete. This represents a big issue.")
		for s in self._solver_list:
			s.finalize()

	def branch(self):
		c = CachingFrontend.branch(self)

		for s in self._solver_list:
			c_s = s.branch()
			for v in c_s.variables:
				c._solvers[v] = c_s #pylint:disable=no-member

		if self.result is not None:
			c.result = self.result.branch()

		return c

	@property
	def timeout(self):
		return self._template_frontend.timeout

	@timeout.setter
	def timeout(self, t):
		self._template_frontend.timeout = t
		for s in self._solver_list:
			s.timeout = t

	# this is for backwards compatibility
	@property
	def _solver_backend(self):
		return self._template_frontend._solver_backend

	def _blank_copy(self):
		return CompositeFrontend(self._template_frontend, cache=self._cache)

	def merge(self, others, merge_flag, merge_values):
		l.debug("Merging %s with %d other solvers.", self, len(others))
		merged = self._blank_copy()
		common_solvers = self._shared_solvers(others)
		common_ids = { s.uuid for s in common_solvers }
		l.debug("... %s common solvers", len(common_solvers))

		for s in common_solvers:
			for v in s.variables:
				merged._solvers[v] = s.branch()

		noncommon_solvers = [ [ s for s in cs._solver_list if s.uuid not in common_ids ] for cs in [self]+others ]

		l.debug("... merging noncommon solvers")
		combined_noncommons = [ ]
		for ns in noncommon_solvers:
			l.debug("... %d", len(ns))
			if len(ns) == 0:
				s = self._blank_copy()
				s.add(True)
				combined_noncommons.append(s)
			elif len(ns) == 1:
				combined_noncommons.append(ns[0])
			else:
				combined_noncommons.append(ns[0].combine(ns[1:]))

		_, merged_noncommon = combined_noncommons[0].merge(combined_noncommons[1:], merge_flag, merge_values)
		for v in merged_noncommon.variables:
			merged._solvers[v] = merged_noncommon

		return True, merged

	def combine(self, others):
		combined = self._blank_copy()
		combined._simplified = False
		combined.add(self.constraints)
		for o in others:
			combined.add(o.constraints)
		return combined

	def split(self):
		return [ s.branch() for s in self._solver_list if len(s.variables) > 0 ]

from ..result import Result
from ..errors import ClaripyFrontendError



from .bits import Bits
from ..ast.base import _make_name

_bvv_cache = dict()

# This is a hilarious hack to get around some sort of bug in z3's python bindings, where
# under some circumstances stuff gets destructed out of order
def cleanup():
    global _bvv_cache #pylint:disable=global-variable-not-assigned
    del _bvv_cache
import atexit
atexit.register(cleanup)

class BV(Bits):

    # TODO: do these go on Bits or BV?
    def chop(self, bits=1):
        """
        Chops an AST into ASTs of size 'bits'. Obviously, the length of the AST must be
        a multiple of bits.
        """
        s = len(self)
        if s % bits != 0:
            raise ValueError("expression length (%d) should be a multiple of 'bits' (%d)" % (len(self), bits))
        elif s == bits:
            return [ self ]
        else:
            return list(reversed([ self[(n+1)*bits - 1:n*bits] for n in range(0, s / bits) ]))

    def __getitem__(self, rng):
        """
        Extracts bits from the AST. ASTs are indexed weirdly. For a 32-bit AST:

            a[31] is the *LEFT* most bit, so it'd be the 0 in

                01111111111111111111111111111111

            a[0] is the *RIGHT* most bit, so it'd be the 0 in

                11111111111111111111111111111110

            a[31:30] are the two leftmost bits, so they'd be the 0s in:

                00111111111111111111111111111111

            a[1:0] are the two rightmost bits, so they'd be the 0s in:

                11111111111111111111111111111100

        :return: the new AST.
        """
        if type(rng) is slice:
            left = rng.start if rng.start is not None else len(self)-1
            right = rng.stop if rng.stop is not None else 0
            if left < 0:
                left = len(self) + left
            if right < 0:
                right = len(self) + right
            return Extract(left, right, self)
        else:
            return Extract(int(rng), int(rng), self)

    def get_byte(self, index):
        """
        Extracts a byte from a BV, where the index refers to the byte in a big-endian order
        :param index: the byte to extract
        :return:
        """
        pos = self.size() / 8 - 1 - index
        return self[pos * 8 + 7 : pos * 8]

    def zero_extend(self, n):
        """
        Zero-extends the AST by n bits. So:

            a = BVV(0b1111, 4)
            b = a.zero_extend(4)
            b is BVV(0b00001111)
        """
        return ZeroExt(n, self)

    def sign_extend(self, n):
        """
        Sign-extends the AST by n bits. So:

            a = BVV(0b1111, 4)
            b = a.sign_extend(4)
            b is BVV(0b11111111)
        """
        return SignExt(n, self)

    def concat(self, *args):
        """
        Concatenates this AST with the ASTs provided.
        """
        return Concat(self, *args)

    @staticmethod
    def _from_int(like, value):
        return BVV(value, like.length)

    @staticmethod
    def _from_long(like, value):
        return BVV(value, like.length)

    @staticmethod
    def _from_str(like, value): #pylint:disable=unused-argument
        return BVV(value)

    @staticmethod
    def _from_BVV(like, value): #pylint:disable=unused-argument
        return BVV(value.value, value.size())

    def signed_to_fp(self, rm, sort):
        if rm is None:
            rm = fp.fp.RM.default()

        return fp.fpToFP(rm, self, sort)

    def unsigned_to_fp(self, rm, sort):
        if rm is None:
            rm = fp.fp.RM.default()
        return fp.fpToFPUnsigned(rm, self, sort)

    def raw_to_fp(self):
        sort = fp.fp.FSort.from_size(self.length)
        return fp.fpToFP(self, sort)

    def to_bv(self):
        return self

def BVS(name, size, min=None, max=None, stride=None, uninitialized=False,  #pylint:disable=redefined-builtin
        explicit_name=None, discrete_set=False, discrete_set_max_card=None, **kwargs):
    """
    Creates a bit-vector symbol (i.e., a variable).

    :param name:            The name of the symbol.
    :param size:            The size (in bits) of the bit-vector.
    :param min:             The minimum value of the symbol.
    :param max:             The maximum value of the symbol.
    :param stride:          The stride of the symbol.
    :param uninitialized:   Whether this value should be counted as an "uninitialized" value in the course of an
                            analysis.
    :param bool explicit_name:   If False, an identifier is appended to the name to ensure uniqueness.
    :param bool discrete_set: If True, a DiscreteStridedIntervalSet will be used instead of a normal StridedInterval.
    :param int discrete_set_max_card: The maximum cardinality of the discrete set. It is ignored if discrete_set is set
                                      to False or None.

    :returns:               a BV object representing this symbol.
    """

    if stride == 0 and max != min:
        raise ClaripyValueError("BVSes of stride 0 should have max == min")

    n = _make_name(name, size, False if explicit_name is None else explicit_name)

    if not discrete_set:
        discrete_set_max_card = None

    return BV('BVS', (n, min, max, stride, uninitialized, discrete_set, discrete_set_max_card), variables={n},
              length=size, symbolic=True, eager_backends=None, uninitialized=uninitialized, **kwargs)

def BVV(value, size=None, **kwargs):
    """
    Creates a bit-vector value (i.e., a concrete value).

    :param value:   The value.
    :param size:    The size (in bits) of the bit-vector.

    :returns:       A BV object representing this value.
    """

    if type(value) is str:
        if size is None:
            size = 8*len(value)
            value = int(value.encode('hex'), 16) if value != "" else 0
        elif size == len(value)*8:
            value = int(value.encode('hex'), 16) if value != "" else 0
        else:
            raise ClaripyValueError('string/size mismatch for BVV creation')
    elif size is None:
        raise ClaripyValueError('BVV() takes either an integer value and a size or a string of bytes')

    if not kwargs:
        try: return _bvv_cache[(value, size)]
        except KeyError: pass

    result = BV('BVV', (value, size), length=size, **kwargs)
    _bvv_cache[(value, size)] = result
    return result

def SI(name=None, bits=0, lower_bound=None, upper_bound=None, stride=None, to_conv=None, explicit_name=None,
       discrete_set=False, discrete_set_max_card=None):
    name = 'unnamed' if name is None else name
    if to_conv is not None:
        si = vsa.CreateStridedInterval(name=name, bits=bits, lower_bound=lower_bound, upper_bound=upper_bound, stride=stride, to_conv=to_conv)
        return BVS(name, si._bits, min=si._lower_bound, max=si._upper_bound, stride=si._stride, explicit_name=explicit_name)
    return BVS(name, bits, min=lower_bound, max=upper_bound, stride=stride, explicit_name=explicit_name,
               discrete_set=discrete_set, discrete_set_max_card=discrete_set_max_card)

def TSI(bits, name=None, uninitialized=False, explicit_name=None):
    name = 'unnamed' if name is None else name
    return BVS(name, bits, uninitialized=uninitialized, explicit_name=explicit_name)

def ESI(bits, **kwargs):
    return BVV(None, bits, **kwargs)

def ValueSet(**kwargs):
    vs = vsa.ValueSet(**kwargs)
    return BV('I', (vs,), variables={ vs.name }, symbolic=False, length=kwargs['bits'], eager_backends=None)
VS = ValueSet

def DSIS(name=None, bits=0, lower_bound=None, upper_bound=None, stride=None, explicit_name=None, to_conv=None, max_card=None):

    if to_conv is not None:
        si = vsa.CreateStridedInterval(bits=to_conv.size(), to_conv=to_conv)
        return SI(name=name, bits=si._bits, lower_bound=si._lower_bound, upper_bound=si._upper_bound, stride=si._stride,
                   explicit_name=explicit_name, discrete_set=True, discrete_set_max_card=max_card)
    else:
        return SI(name=name, bits=bits, lower_bound=lower_bound, upper_bound=upper_bound, stride=stride,
                   explicit_name=explicit_name, discrete_set=True, discrete_set_max_card=max_card)

#
# Unbound operations
#

from .bool import Bool
from .. import operations

# comparisons
ULT = operations.op('__lt__', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
ULE = operations.op('__le__', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
UGT = operations.op('__gt__', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
UGE = operations.op('__ge__', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
SLT = operations.op('SLT', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
SLE = operations.op('SLE', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
SGT = operations.op('SGT', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)
SGE = operations.op('SGE', (BV, BV), Bool, extra_check=operations.length_same_check, bound=False)

# bit stuff
LShR = operations.op('LShR', (BV, BV), BV, extra_check=operations.length_same_check,
                     calc_length=operations.basic_length_calc, bound=False)
SignExt = operations.op('SignExt', ((int, long), BV), BV,
                        calc_length=operations.ext_length_calc, bound=False)
ZeroExt = operations.op('ZeroExt', ((int, long), BV), BV,
                        calc_length=operations.ext_length_calc, bound=False)
Extract = operations.op('Extract', ((int, long), (int, long), BV),
                        BV, extra_check=operations.extract_check,
                        calc_length=operations.extract_length_calc, bound=False)

Concat = operations.op('Concat', BV, BV, calc_length=operations.concat_length_calc, bound=False)

RotateLeft = operations.op('RotateLeft', (BV, BV), BV,
                           extra_check=operations.length_same_check,
                           calc_length=operations.basic_length_calc, bound=False)
RotateRight = operations.op('RotateRight', (BV, BV), BV,
                            extra_check=operations.length_same_check,
                            calc_length=operations.basic_length_calc, bound=False)
Reverse = operations.op('Reverse', (BV,), BV,
                        calc_length=operations.basic_length_calc, bound=False)

union = operations.op('union', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc, bound=False)
widen = operations.op('widen', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc, bound=False)
intersection = operations.op('intersection', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc, bound=False)

#
# Bound operations
#

BV.__add__ = operations.op('__add__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__radd__ = operations.reversed_op(BV.__add__.im_func)
BV.__div__ = operations.op('__div__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rdiv__ = operations.reversed_op(BV.__div__.im_func)
BV.__truediv__ = operations.op('__truediv__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rtruediv__ = operations.reversed_op(BV.__truediv__.im_func)
BV.__floordiv__ = operations.op('__floordiv__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rfloordiv__ = operations.reversed_op(BV.__floordiv__.im_func)
BV.__mul__ = operations.op('__mul__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rmul__ = operations.reversed_op(BV.__mul__.im_func)
BV.__sub__ = operations.op('__sub__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rsub__ = operations.reversed_op(BV.__sub__.im_func)
BV.__pow__ = operations.op('__pow__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rpow__ = operations.reversed_op(BV.__pow__.im_func)
BV.__mod__ = operations.op('__mod__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rmod__ = operations.reversed_op(BV.__mod__.im_func)
BV.__divmod__ = operations.op('__divmod__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rdivmod__ = operations.reversed_op(BV.__divmod__.im_func)

BV.__neg__ = operations.op('__neg__', (BV,), BV, calc_length=operations.basic_length_calc)
BV.__pos__ = operations.op('__pos__', (BV,), BV, calc_length=operations.basic_length_calc)
BV.__abs__ = operations.op('__abs__', (BV,), BV, calc_length=operations.basic_length_calc)

BV.__eq__ = operations.op('__eq__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__ne__ = operations.op('__ne__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__ge__ = operations.op('__ge__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__le__ = operations.op('__le__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__gt__ = operations.op('__gt__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.__lt__ = operations.op('__lt__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SLT = operations.op('SLT', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SGT = operations.op('SGT', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SLE = operations.op('SLE', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.SGE = operations.op('SGE', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.ULT = operations.op('__lt__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.UGT = operations.op('__gt__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.ULE = operations.op('__le__', (BV, BV), Bool, extra_check=operations.length_same_check)
BV.UGE = operations.op('__ge__', (BV, BV), Bool, extra_check=operations.length_same_check)

BV.__invert__ = operations.op('__invert__', (BV,), BV, calc_length=operations.basic_length_calc)
BV.__or__ = operations.op('__or__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__ror__ = operations.reversed_op(BV.__or__.im_func)
BV.__and__ = operations.op('__and__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rand__ = operations.reversed_op(BV.__and__.im_func)
BV.__xor__ = operations.op('__xor__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rxor__ = operations.reversed_op(BV.__xor__.im_func)
BV.__lshift__ = operations.op('__lshift__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rlshift__ = operations.reversed_op(BV.__lshift__.im_func)
BV.__rshift__ = operations.op('__rshift__', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.__rrshift__ = operations.reversed_op(BV.__rshift__.im_func)
BV.LShR = operations.op('LShR', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)

BV.Extract = staticmethod(operations.op('Extract', ((int, long), (int, long), BV), BV, extra_check=operations.extract_check, calc_length=operations.extract_length_calc, bound=False))
BV.Concat = staticmethod(operations.op('Concat', BV, BV, calc_length=operations.concat_length_calc, bound=False))
BV.reversed = property(operations.op('Reverse', (BV,), BV, calc_length=operations.basic_length_calc))

BV.union = operations.op('union', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.widen = operations.op('widen', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)
BV.intersection = operations.op('intersection', (BV, BV), BV, extra_check=operations.length_same_check, calc_length=operations.basic_length_calc)

from . import fp
from .. import vsa
from ..errors import ClaripyValueError
